/**
  *  \file planetscentral/host/ranking.cc
  *  \brief Host: Player Ranking
  */

#include <cstdio>
#include "planetscentral/host/ranking.h"
#include "planetscentral/dbclient/dbclient.h"
#include "planetscentral/dbclient/dbhash.h"
#include "planetscentral/dbclient/dbsubtree.h"
#include "planetscentral/host/db.h"
#include "cpluslib/misc.h"
#include "planetscentral/dbclient/reconnect.h"
#include "planetscentral/host/host.h"
#include "planetscentral/dbclient/mailclient.h"
#include "io/io.h"

static const int32_t RELIABILITY_SCALE = 1000;

static const int32_t RELIABILITY_SPEED = 3;

static const int32_t DROP_PENALTY = 66;

//      1.          Admiral                    50000    90  >95    500       15            [Admiral]
//      2.          Vice Admiral               25000    90  >95    400       14            [Vizeadmiral]
//      3.          Rear Admiral               16000    90  >95    300       13            [Konteradmiral]
//      4.          Captain                    12000    80  >85    300       12            [Kapitn [zur See]]
//      5.          Commander                  9000     80  >85    250       11            [Fregattenkapitn]
//      6.          Lieutenant Commander       6500     80  >85    200       10            [~Korvettenkapitn]
//      7.          Lieutenant                 4000     70  >75    200       9             [Kapitnleutnant]
//      8.          Lieutenant Junior Grade    2500     70  >75    150       8             [Oberleutnant zur See]
//      9.          Ensign                     1750     70  >75    100       7             [Leutnant zur See, ~Fhnrich]
//      10.         Senior Chief Petty Officer 1200     60  >65    100       6             [~Stabsbootsmann]
//      11.         Chief Petty Officer        800      60  >65    80        5             [~Oberbootsmann]
//      12.         Petty Officer              500      60  >65    60        4             [Bootsmann]
//      13.         Spaceman                   250      50  >55    40        3             [Raumflieger]
//      14.         Spaceman Apprentice        100      50  >55    20        2             [Raumflieger-Anwrter]
//      15.         Spaceman Recruit           0        0          0         1             [Raumflieger-Rekrut]

namespace {
    struct RankDefinition {
        int32_t min_rankpoints;            // Minimum rankpoints to get this rank
        int16_t min_turnreliability;       // Minimum turnreliability to get this rank
        int16_t min_turnreliability_keep;  // Minimum turnreliability to keep this rank
        int32_t min_turns;                 // Minimum turns played to get this rank
    };

    const RankDefinition rank_definitions[] = {
        {   100, 55, 50,  20 },    // Spaceman Apprentice
        {   250, 55, 50,  40 },    // Spaceman
        {   500, 65, 60,  60 },    // Petty Officer
        {   800, 65, 60,  80 },    // Chief Petty Officer
        {  1200, 65, 60, 100 },    // Senior Chief Petty Officer
        {  1750, 75, 70, 100 },    // Ensign
        {  2500, 75, 70, 150 },    // Lieutenant Junior Grade
        {  4000, 75, 70, 200 },    // Lieutenant
        {  6500, 85, 80, 200 },    // Lieutenant Commander
        {  9000, 85, 80, 250 },    // Commander
        { 12000, 85, 80, 300 },    // Captain
        { 16000, 95, 90, 300 },    // Rear Admiral
        { 25000, 95, 90, 400 },    // Vice Admiral
        { 50000, 95, 90, 500 },    // Admiral
    };

    static const int32_t MAX_RANK = countof(rank_definitions);
}

/** Handle player turn submission.
    \param dbConnection Database; locked
    \param player Player number
    \param submit true iff turn was submitted
    \param level Replacement level (0=current player in charge of slot) */
void
handlePlayerTurn(DbClient& dbConnection, string_t player, bool submit, uint32_t level)
{
    DbHash profile(DbSubtree(dbConnection, USER_ROOT).subtree(player).hashKey("profile"));

    // Count this turn
    if (submit)
        ++profile.intField("turnsplayed");
    else
        ++profile.intField("turnsmissed");

    // Adjust reliability
    int32_t newPoints = RELIABILITY_SCALE * RELIABILITY_SPEED;
    if (!submit && level < 30) {
        newPoints -= newPoints >> level;
    }
    profile.setField("turnreliability", profile.getIntField("turnreliability") * (100 - RELIABILITY_SPEED) / 100 + newPoints);

    // printf("[ranking] player '%s': %d points (%s, level %d)\n", player.c_str(), newPoints, submit?"submit":"miss", int(level));
}

void
handlePlayerDrop(DbClient& dbConnection, string_t player, Game& game, int slot)
{
    // The game must have started; otherwise, we cannot give a penalty.
    int32_t turn = game.getConfigInt("turn");
    if (turn <= 0) {
        return;
    }

    // Get current turn's scores
    string_t packedScores = game.getGame().subtree("turn").subtree(turn).hashKey("scores").getStringField(game.getRefereeScoreName());

    // Do we actually have scores for this player?
    if (packedScores.size() < slot*4U) {
        return;
    }
    int32_t playerScore = getInt32(packedScores.data() + 4*(slot-1));

    // Find maximum score
    int32_t maxScore = -1;
    for (uint_t i = 1; i*4 <= packedScores.size(); ++i) {
        int32_t thisScore = getInt32(packedScores.data() + 4*(i-1));
        if (thisScore > maxScore) {
            maxScore = thisScore;
        }
    }

    // Can we compute a penalty?
    if (playerScore < 0 || maxScore <= 0) {
        return;
    }

    // Give penalty
    DbHash profile(DbSubtree(dbConnection, USER_ROOT).subtree(player).hashKey("profile"));
    profile.setField("turnreliability", int32_t(profile.getIntField("turnreliability") * double(maxScore*100 - playerScore*DROP_PENALTY) / (maxScore*100)));
}

void
addPlayerRankPoints(DbClient& dbConnection, string_t player, int32_t pts)
{
    DbHash profile(DbSubtree(dbConnection, USER_ROOT).subtree(player).hashKey("profile"));
    profile.intField("rankpoints") += pts;
}

/** Check whether this player needs a rank change. If so, performs it. */
void
handlePlayerRankChanges(DbClient& dbConnection, string_t player)
{
    DbHash profile(DbSubtree(dbConnection, USER_ROOT).subtree(player).hashKey("profile"));

    int32_t currentRank        = profile.getIntField("rank");
    int32_t currentRankPoints  = profile.getIntField("rankpoints");
    int32_t currentReliability = profile.getIntField("turnreliability");
    int32_t currentTurns       = profile.getIntField("turnsplayed");

    int32_t roundedReliability = (currentReliability + RELIABILITY_SCALE/2) / RELIABILITY_SCALE;

    const char* mail = 0;
    while (currentRank < MAX_RANK
        && currentRankPoints >= rank_definitions[currentRank].min_rankpoints
        && roundedReliability >= rank_definitions[currentRank].min_turnreliability
        && currentTurns >= rank_definitions[currentRank].min_turns)
    {
        // promotion
        ++currentRank;
        profile.setField("rank", currentRank);
        mail = "rank-promotion";
    }

    while (currentRank > 0
           && (currentRankPoints < rank_definitions[currentRank-1].min_rankpoints
               || roundedReliability < rank_definitions[currentRank-1].min_turnreliability_keep
               || currentTurns < rank_definitions[currentRank-1].min_turns))
    {
        // demotion
        --currentRank;
        profile.setField("rank", currentRank);
        mail = "rank-demotion";
    }

    // So, did something happen?
    if (mail != 0) {
        std::printf("[ranking] %s for user %s, new rank: %d\n", mail, player.c_str(), currentRank);
        ReconnectingGuard g(mail_rc);
        mail_connection.startMail(mail, "rank-" + player);
        mail_connection.addParameter("rank", currentRank);
        mail_connection.addParameter("rankpoints", currentRankPoints);
        mail_connection.addParameter("turnreliability", roundedReliability);
        mail_connection.addParameter("turnsplayed", currentTurns);
        mail_connection.sendTo("user:" + player);
    }
}

