/**
  *  \file planetscentral/host/victory.cc
  *  \brief Host: Victory Check and Victory Point Computation
  */

#include <cstring>
#include <cstdio>
#include <map>
#include "planetscentral/host/victory.h"
#include "cpluslib/format.h"
#include "cpluslib/misc.h"
#include "game/playerset.h"
#include "game/struct.h"
#include "int/array.h"
#include "int/value.h"
#include "io/dirs.h"
#include "io/io.h"
#include "io/streams.h"
#include "io/textfile.h"
#include "planetscentral/host/db.h"
#include "planetscentral/host/game.h"
#include "planetscentral/host/rank.h"
#include "planetscentral/host/ranking.h"
#include "util/misc.h"
#include "util/random.h"

/******************************* Utilities *******************************/

/** Save ranks in game.
    \param ranks [in] Ranks, not necessarily compacted
    \param game  Game to store to */
static void
saveRanks(const Rank_t& ranks, Game& game)
{
    /* Compact ranks using a null tiebreaker */
    GPlayerSet slots = game.getGameSlots();
    Rank_t result;
    Rank_t null;
    initRanks(null);
    compactRanks(result, ranks, null, slots);

    /* Save */
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        if (slots.contains(i)) {
            game.getSlot(i).rank().set(result[i-1]);
        } else {
            game.getSlot(i).rank().set(0);
        }
    }
}

/** Determine ranks from score, and save them. */
static void
saveScoreRanks(Game& game)
{
    Rank_t ranks;
    initRanks(ranks);

    /* Fetch all scores, but negate them because a high score is a good rank */
    const int32_t turn = game.getConfigInt("turn");
    const string_t scoreName   = game.getRefereeScoreName();
    const string_t scoreRecord = game.getGame().subtree("turn").subtree(turn).hashKey("scores").getStringField(scoreName);
    for (uint_t i = 1; i <= NUM_PLAYERS; ++i) {
        if (scoreRecord.size() >= 4*i) {
            ranks[i-1] = -getInt32(scoreRecord.data() + 4*(i-1));
        }
    }

    /* Save */
    saveRanks(ranks, game);
}

/** Compute data for a score condition.
    \param game     Game to work on
    \param fromTurn [in] Start in this turn, inclusive
    \param toTurn   [in] End in this turn, inclusive
    \param players  [in] Players in game
    \param turns    [out] Negated number of turns-over-limit
    \param scores   [out] Final turn's scores */
static void
computeScoreCondition(Game& game,
                      const int fromTurn,
                      const int toTurn,
                      const GPlayerSet players,
                      Rank_t& turns,
                      Rank_t& scores)
{
    /* Get score */
    const int32_t endScore = game.getConfigInt("endScore");
    const string_t endScoreName = game.getRefereeScoreName();

    /* Parse it. Fetch negated turns, so that it can be used as a rank structure */
    initRanks(turns, 0);
    initRanks(scores, 0);
    for (int i = fromTurn; i <= toTurn; ++i) {
        /* Fetch scores */
        string_t scoreRecord = game.getGame().subtree("turn").subtree(i).hashKey("scores").getStringField(endScoreName);
        uint_t count = scoreRecord.size() / 4;
        if (count > NUM_PLAYERS) {
            count = NUM_PLAYERS;
        }
        getInt32(scoreRecord.data(), scores, count);

        /* Compare */
        for (int pl = 1; pl <= NUM_PLAYERS; ++pl) {
            if (players.contains(pl)) {
                if (scores[pl-1] >= endScore) {
                    --turns[pl-1];
                } else {
                    turns[pl-1] = 0;
                }
            }
        }
    }
}

/*************************** Victory Conditions **************************/

/** Check "turn" condition. Game ends after reaching a particular turn. */
static bool
checkTurnCondition(string_t gameDir, Game& game)
{
    /* Check turn number. Game does not end if endTurn not reached */
    int32_t turn = game.getConfigInt("turn");
    int32_t endTurn = game.getConfigInt("endTurn");
    if (turn < endTurn) {
        return false;
    }

    /* Check endProbability. 0 (unset) or 100 means we don't roll a die. */
    int32_t endProbability = game.getConfigInt("endProbability");
    int32_t effProbability = endProbability * (2 + turn - endTurn) / 2;
    if (effProbability <= 0
        || effProbability >= 100
        || myRandom(100) < effProbability)
    {
        /* Game ends */
        std::printf("[cron] %s: 'turn' condition ends game\n", gameDir.c_str());
        saveScoreRanks(game);
        return true;
    }

    /* Keep going */
    std::printf("[cron] %s: 'turn' condition continues game, probability %d\n", gameDir.c_str(), int(effProbability));
    return false;
}

/** Check "score" condition. Game ends when someone reaches a particular score. */
static bool
checkScoreCondition(string_t gameDir, Game& game)
{
    /* Check turn number. If endTurn is not yet reached, nobody can possibly have enough points. */
    int32_t turn = game.getConfigInt("turn");
    int32_t endTurn = game.getConfigInt("endTurn");
    if (endTurn < 1) {
        endTurn = 1;
    }
    if (turn < endTurn) {
        return false;
    }

    /* Get score */
    Rank_t turns;
    Rank_t scores;
    const GPlayerSet players = game.getGameSlots();
    computeScoreCondition(game, turn - endTurn + 1, turn, players, turns, scores);

    /* Do we have a winner? */
    bool end = false;
    for (int pl = 1; pl <= NUM_PLAYERS; ++pl) {
        if (-turns[pl-1] >= endTurn) {
            end = true;
        }
    }
    if (!end) {
        return false;
    }
    std::printf("[cron] %s: 'score' condition ends game\n", gameDir.c_str());

    /* Build ranks */
    Rank_t ranks;
    compactRanks(ranks, turns, scores, players);

    /* Save it and end game */
    saveRanks(ranks, game);
    return true;
}

/** Check victory condition.
    \param gameDir Physical game directory
    \param game    Game to check; locked

    \retval false Game continues, no changes made
    \retval true  Game ends. Ranks have been saved into database. Caller must mark game finished,
                  send mail, and process ranking system. */
bool
checkVictory(string_t gameDir, Game& game)
{
    /* Read plug-in victory status */
    Ptr<Stream> ref = Stream::openFileNT(Directory::makeFileName(gameDir, "c2ref.txt"), Stream::C_READ);
    if (ref) {
        TextFile tf(*ref);
        RefereeFileReader rdr;
        string_t line;
        while (tf.getLine(line)) {
            rdr.processLine(line);
        }
        ref = 0;
        if (rdr.end) {
            std::printf("[cron] %s: game end declared by add-on\n", gameDir.c_str());
            saveRanks(rdr.ranks, game);
            return true;
        }
    }

    /* Check builtin victory conditions */
    string_t cond = game.getConfig("endCondition");
    if (cond == "turn") {
        /* End after turn, with probability */
        return checkTurnCondition(gameDir, game);
    } else if (cond == "score") {
        /* End when score reached (and kept for some turns) */
        return checkScoreCondition(gameDir, game);
    }
    return false;
}

/** Check victory for forced game end. If a game is forcibly terminated, this tries to
    make up a rank list.
    \param gameDir Physical game directory
    \param game    Game to check; locked */
void
checkForcedGameEnd(Game& game)
{
    int32_t turn = game.getConfigInt("turn");
    if (turn <= 0) {
        /* There are no scores. Treat everyone equal. */
        Rank_t null;
        initRanks(null);
        saveRanks(null, game);
    } else if (game.getConfig("endCondition") == "score") {
        /* This must approximate the actual game score. If a "120 planets for 5 turns"
           game is terminated, a player with 120 planets for 4 turns is closer to winning than
           a player with 150 planets for 1 turn. Thus, we count how many turns over limit
           players are. */
        int32_t endTurn = game.getConfigInt("endTurn");
        if (endTurn < 1) {
            endTurn = 1;
        }
        int32_t firstTurnToJudge = turn - endTurn + 1;
        if (firstTurnToJudge <= 0) {
            firstTurnToJudge = 1;
        }

        /* Compute partial score */
        Rank_t turns;
        Rank_t scores;
        const GPlayerSet players = game.getGameSlots();
        computeScoreCondition(game, firstTurnToJudge, turn, players, turns, scores);

        /* Save */
        Rank_t ranks;
        compactRanks(ranks, turns, scores, players);
        saveRanks(ranks, game);
    } else {
        /* All other scoring modes just use the scores. This is the exact result for "turn"
           condition, and a very rough approximate for add-on conditions (if we play a game
           judged by an add-on, we cannot know how close to winning a player is. */
        saveScoreRanks(game);
    }
}


namespace {
    struct PlayerInfo {
        int32_t turnsTotal;
        int32_t turnsPlayed;
        int32_t rankLevel;
        int32_t turnsInSlot[NUM_PLAYERS];

        PlayerInfo()
            : turnsTotal(0), turnsPlayed(0), rankLevel(0)
            {
                for (uint_t i = 0; i < countof(turnsInSlot); ++i) {
                    turnsInSlot[i] = 0;
                }
            }
    };

    static const int16_t POSITION_POINTS[] = { 20, 14, 10, 8, 6, 4, 3, 2 };
};

/** Compute rankings for a game. */
void
computeGameRankings(DbClient& dbConnection, Game& game)
{
    DbSubtree userRoot(dbConnection, USER_ROOT);
    DbHash rankPointsHash(game.getGame().hashKey("rankpoints"));
    int32_t currentTurn = game.getConfigInt("turn");
    // int32_t existingTurn = game.getConfigInt("rankTurn");   // we do not use that yet
    std::map<string_t, PlayerInfo> players;

    Rank_t sumOfRankLevels;
    Rank_t lastTurnSeen;
    initRanks(sumOfRankLevels, 0);
    initRanks(lastTurnSeen, 0);

    // Scan through whole game and collect information.
    // Start at turn 2, because everyone "misses" turn 1. This means players who drop in
    // turn one are not registered in this game. Which is not surprising, because players
    // that join and immediately resign are not counted either.
    std::printf("[ranking] Computing ranks for game %d, %d turns...\n", int(game.getId()), int(currentTurn));
    for (int32_t turnNr = 2; turnNr <= currentTurn; ++turnNr) {
        // Process one turn
        DbSubtree turn(game.getGame().subtree("turn").subtree(turnNr));
        string_t turnStatus = turn.hashKey("info").getStringField("turnstatus");
        for (int slot = 1; slot <= NUM_PLAYERS; ++slot) {
            // Do we have information about this slot?
            if (turnStatus.size() < 2U*slot) {
                continue;
            }

            // Was this slot in the game at this time?
            int32_t slotTurnStatus = getInt16(turnStatus.data() + 2*(slot-1));
            if (slotTurnStatus < 0) {
                continue;
            }
            slotTurnStatus &= TurnStateMask;

            // Was anyone playing this slot?
            string_t player = turn.hashKey("player").getStringField(format("%d", slot));
            if (player.empty()) {
                continue;
            }

            // Locate this player
            std::map<string_t, PlayerInfo>::iterator it = players.find(player);
            if (it == players.end()) {
                it = players.insert(std::make_pair(player, PlayerInfo())).first;
                it->second.rankLevel = userRoot.subtree(player).hashKey("profile").getIntField("rank") + 1;
            }

            // Count slot
            sumOfRankLevels[slot-1] += it->second.rankLevel;
            lastTurnSeen[slot-1] = turnNr;

            // Count player
            if (slotTurnStatus == TurnYellow || slotTurnStatus == TurnGreen || slotTurnStatus == TurnDead) {
                ++it->second.turnsPlayed;
            }
            ++it->second.turnsTotal;
            ++it->second.turnsInSlot[slot-1];
        }
    }

    // Fetch ranks computed by referee
    Rank_t refRanks;
    initRanks(refRanks);
    for (int slot = 1; slot <= NUM_PLAYERS; ++slot) {
        int32_t value = game.getSlot(slot).rank().get();
        if (value) {
            refRanks[slot-1] = value;
        }
    }

    // Compute final ranks
    Rank_t finalRanks;
    compactRanks(finalRanks, refRanks, lastTurnSeen, GPlayerSet::allUpTo(NUM_PLAYERS));

    // Compute game-dependant weighing factor (Game_Difficulty * Turn_Factor)
    double gameFactor = game.getDifficulty() / 100.0;
    if (currentTurn < 50) {
        gameFactor *= currentTurn;
        gameFactor /= 50;
    }

    // Compute scores for all players
    for (std::map<string_t, PlayerInfo>::iterator it = players.begin(); it != players.end(); ++it) {
        int usedSlot = 0;
        int32_t newPoints = 0;
        for (int slot = 1; slot <= NUM_PLAYERS; ++slot) {
            if (it->second.turnsInSlot[slot-1] != 0) {
                // Player was active in this slot, compute score
                const int slotRank = finalRanks[slot-1];
                const int positionPoints = slotRank <= int(countof(POSITION_POINTS)) ? POSITION_POINTS[slotRank-1] : 1;

                double reliabilityRate = double(it->second.turnsPlayed) / it->second.turnsTotal;
                reliabilityRate *= reliabilityRate;

                // Player_Rate_Num
                int numOpponents = 0;
                int32_t sumRanksOfOpponents = 0;
                for (int i = 1; i <= NUM_PLAYERS; ++i) {
                    if (finalRanks[i-1] > slotRank) {
                        ++numOpponents;
                        sumRanksOfOpponents += sumOfRankLevels[i-1];
                    }
                }
                const double playerRateNum = (sumRanksOfOpponents / double(currentTurn)) + 110;

                // Player_Rate_Den. Note that dividing by turnsInSlot automatically
                // scales the value down for players that did not play a whole game.
                const double playerRateDen = (it->second.rankLevel
                                              * numOpponents
                                              + 110)
                    * double(currentTurn)
                    / it->second.turnsInSlot[slot-1];

                // Compute totals. Player gets maximum of totals (not sum).
                const int32_t points = int32_t(100 * gameFactor * positionPoints * playerRateNum / playerRateDen * reliabilityRate + 0.5);
                if (points > newPoints) {
                    newPoints = points;
                    usedSlot = slot;
                }
            }
        }

        // Log and store in database
        const string_t& user = it->first;
        DbHashIntField points(rankPointsHash, user);
        int32_t oldPoints = points.get();
        printf("[ranking]   slot %2d, %5d points (was %5d), user %s\n",
               usedSlot,
               int(newPoints),
               int(oldPoints),
               it->first.c_str());
        points.set(newPoints);
        addPlayerRankPoints(dbConnection, user, newPoints - oldPoints);
    }

    // Trigger rank changes
    for (std::map<string_t, PlayerInfo>::iterator it = players.begin(); it != players.end(); ++it) {
        handlePlayerRankChanges(dbConnection, it->first);
    }
    game.setConfigInt("rankTurn", currentTurn);
    
    std::printf("[ranking] done.\n");
}



/** Describe victory condition for external interface.
    \param game Game, locked
    \param root Database root
    \return newly-allocated result object */
IntValue*
describeVictoryCondition(Game& game, DbClient& root)
{
    Ptr<IntArrayData> ad = new IntArrayData();

    /* Condition */
    string_t cond = game.getConfig("endCondition");
    ad->content.pushNew(makeStringValue("endCondition"));
    ad->content.pushNew(makeStringValue(cond));

    if (cond == "turn") {
        /* Report parameters for 'turn' */
        int32_t endTurn = game.getConfigInt("endTurn");
        int32_t endProbability = game.getConfigInt("endProbability");
        if (endProbability <= 0 || endProbability >= 100) {
            endProbability = 100;
        }
        ad->content.pushNew(makeStringValue("endTurn"));
        ad->content.pushNew(makeIntValue(endTurn));
        ad->content.pushNew(makeStringValue("endProbability"));
        ad->content.pushNew(makeIntValue(endProbability));
    } else if (cond == "score") {
        /* Report parameters for 'score' */
        int32_t endTurn = game.getConfigInt("endTurn");
        if (endTurn < 1) {
            endTurn = 1;
        }
        int32_t endScore = game.getConfigInt("endScore");
        string_t endScoreName = game.getRefereeScoreName();
        ad->content.pushNew(makeStringValue("endTurn"));
        ad->content.pushNew(makeIntValue(endTurn));
        ad->content.pushNew(makeStringValue("endScore"));
        ad->content.pushNew(makeIntValue(endScore));
        ad->content.pushNew(makeStringValue("endScoreName"));
        ad->content.pushNew(makeStringValue(endScoreName));
        ad->content.pushNew(makeStringValue("endScoreDescription"));
        if (endScoreName.empty()) {
            ad->content.pushNew(makeStringValue(string_t()));
        } else {
            ad->content.pushNew(makeStringValue(game.getGame().hashKey("scores").getStringField(endScoreName)));
        }
    } else {
        /* End is determined by an add-on. Find it. */
        string_t addonName = game.getGame().hashKey("toolkind").getStringField("referee");
        if (!addonName.empty()) {
            ad->content.pushNew(makeStringValue("referee"));
            ad->content.pushNew(makeStringValue(addonName));
            ad->content.pushNew(makeStringValue("refereeDescription"));
            ad->content.pushNew(makeStringValue(DbSubtree(root, TOOL_PROGRAM_ROOT).subtree("prog").hashKey(addonName).getStringField("description")));
        }
    }
    ad->addDimension(ad->content.size());
    return new IntArray(ad);
}
