/**
  *  \file planetscentral/talk/render.cc
  *  \brief Talk: Rendering
  */

#include <cstring>
#include "planetscentral/talk/render.h"
#include "planetscentral/talk/bbparse.h"
#include "planetscentral/talk/bblex.h"
#include "planetscentral/talk/htmlout.h"
#include "planetscentral/talk/bbout.h"
#include "planetscentral/dbclient/connection.h"
#include "planetscentral/talk/message.h"
#include "planetscentral/talk/user.h"
#include "cpluslib/format.h"
#include "planetscentral/talk/mailout.h"

static void
addParagraph(TextNode* root, const string_t& str)
{
    std::auto_ptr<TextNode> par(new TextNode());
    par->major = maParagraph;
    par->minor = miParNormal;

    std::auto_ptr<TextNode> txt(new TextNode());
    txt->major = maPlain;
    txt->minor = 0;
    txt->text = str;
    par->children.push_back(txt.release());
    root->children.push_back(par.release());
}

static TextNode*
parse(const string_t& str)
{
    if (str.size() >= 6 && str.compare(0, 6, "forum:", 6) == 0) {
        // parse as BBCode
        BBLexer lex(str.substr(6));
        return BBParser(lex).parse();
    } else if (str.size() >= 5 && str.compare(0, 5, "text:", 5) == 0) {
        // parse as text
        std::auto_ptr<TextNode> result(new TextNode());
        result->major = maGroup;
        result->minor = miGroupRoot;

        string_t::size_type i = 5, n;
        while ((n = str.find_first_of("\r\n", i)) != string_t::npos) {
            if (n != i) {
                addParagraph(result.get(), string_t(str, i, n-i));
            }
            i = n+1;
        }
        if (i != str.size() || result->children.empty()) {
            addParagraph(result.get(), string_t(str, i));
        }
        return result.release();
    } else {
        // error, unsupported format.
        std::auto_ptr<TextNode> result(new TextNode());
        result->major = maGroup;
        result->minor = miGroupRoot;
        addParagraph(result.get(), str);
        return result.release();
    }
}

static bool
stripBreak(TextNode* n)
{
    if (n->major == maParagraph && n->minor == miParBreak) {
        return true;
    } else {
        std::size_t i = 0, e = n->children.size();
        while (i < e && !stripBreak(n->children[i])) {
            ++i;
        }
        if (i < e) {
            // There was a break among our children: drop everything after
            ++i;
            while (i < n->children.size()) {
                n->children.pop_back();
            }
            return true;
        } else {
            // No break, keep it
            return false;
        }
    }
}

/****************************** 'abstract:' ******************************/

namespace {
    class Abstract {
     public:
        Abstract();
        bool strip(TextNode* n);
     private:
        uint_t paras;
        uint_t chars;
    };
}

Abstract::Abstract()
    : paras(2),
      chars(200)
{ }

bool
Abstract::strip(TextNode* n)
{
    if (n->major == maParagraph && n->minor == miParBreak) {
        // [break]
        return true;
    } else if ((n->major == maParagraph || n->major == maGroup || n->major == maPlain) && (paras == 0 || chars == 0)) {
        // a paragraph although no more are allowed
        return true;
    } else if (n->major == maPlain) {
        // text
        if (n->text.size() <= chars) {
            chars -= n->text.size();
        } else {
            while (chars > 0 && std::strchr(" \t\r\n", n->text[chars-1]) == 0) {
                --chars;
            }
            n->text.erase(chars);
            n->text += "...";
            chars = 0;
        }
        return false;
    } else {
        // check content
        std::size_t i = 0, e = n->children.size();
        while (i < e && !strip(n->children[i])) {
            ++i;
        }
        if (i < e) {
            // There was a break among our children: drop everything after
            while (i < n->children.size()) {
                n->children.pop_back();
            }
        }

        // count paragraphs
        if (n->major == maParagraph && paras > 0) {
            --paras;
        }

        // No break, keep it
        return false;
    }
}

/****************************** RenderState ******************************/

/** Constructor. */
RenderState::RenderState()
    : baseUrl(),
      format("raw"),
      messageId(0),
      user()
{ }

void
RenderState::clear()
{
    baseUrl.clear();
    format = "raw";
    messageId = 0;
}

void
RenderState::setBaseUrl(string_t url)
{
    this->baseUrl = url;
}

void
RenderState::setFormat(string_t format)
{
    this->format = format;
}

void
RenderState::setMessageId(int32_t id)
{
    this->messageId = id;
}

void
RenderState::setUser(string_t user)
{
    this->user = user;
}

const string_t&
RenderState::getBaseUrl() const
{
    return baseUrl;
}

const string_t&
RenderState::getUser() const
{
    return user;
}

int32_t
RenderState::getMessageId() const
{
    return messageId;
}

bool
RenderState::handleOption(const string_t optUC, IntArgBlock& args)
{
    if (optUC == "FORMAT") {
        args.checkArgumentCountAtLeast(1);
        setFormat(toStringResult(args.getNext()));
        return true;
    } else if (optUC == "BASEURL") {
        args.checkArgumentCountAtLeast(1);
        setBaseUrl(toStringResult(args.getNext()));
        return true;
    } else {
        return false;
    }
}

string_t
RenderState::render(string_t text, DbClient& db)
{
    if (format == "raw") {
        // raw format requested
        return text;
    } else if (format.find(':') == format.npos
               && text.size() > format.size()
               && text.compare(0, format.size(), format) == 0
               && text[format.size()] == ':')
    {
        // requested same format as stored
        return text.substr(format.size()+1);
    } else {
        // transformation required
        std::auto_ptr<TextNode> tree(parse(text));

        // perform conversions
        string_t fmt = format;
        while (1) {
            if (fmt.size() > 6 && fmt.compare(0, 6, "quote:", 6) == 0) {
                // quote: quote entire message
                fmt.erase(0, 6);
                std::auto_ptr<TextNode> tmp(new TextNode());
                tmp->major = maGroup;
                tmp->minor = miGroupQuote;
                if (messageId > 0) {
                    // messageId is trusted, so no permission checks required
                    Message m(db, messageId);
                    User u(db, m.author().get());
                    tmp->text = ::format("%s;%d", u.getLoginName(), messageId);
                }
                tmp->children.swap(tree->children);
                tree->children.push_back(tmp.release());
            } else if (fmt.size() > 8 && fmt.compare(0, 8, "noquote:", 8) == 0) {
                // noquote: remove all quotes
                fmt.erase(0, 8);
                tree->stripQuotes();
            } else if (fmt.size() > 6 && fmt.compare(0, 6, "break:", 6) == 0) {
                // render up to break
                fmt.erase(0, 6);
                stripBreak(tree.get());
            } else if (fmt.size() > 6 && fmt.compare(0, 9, "abstract:", 9) == 0) {
                // render abstract
                fmt.erase(0, 9);
                tree->stripQuotes();
                Abstract().strip(tree.get());
            } else if (fmt.size() > 6 && fmt.compare(0, 6, "force:", 6) == 0) {
                // force: null operation just to force re-rendering
                fmt.erase(0, 6);
            } else {
                break;
            }
        }

        // render it
        if (fmt == "html") {
            return renderHTML(tree.get(), *this, db);
        } else if (fmt == "forum") {
            return renderBB(tree.get());
        } else if (fmt == "mail") {
            return renderMail(tree.get(), false, *this, db);
        } else if (fmt == "news") {
            return renderMail(tree.get(), true, *this, db);
        } else if (fmt == "text") {
            return tree->rawTextContent();
        } else {
            // error
            return "ERROR: invalid format '" + fmt + "'";
        }
    }
}
